use signal_hook::{consts::*, iterator::Signals};
use std::{
    collections::HashMap,
    error::Error,
    fs::{self},
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
    path::PathBuf,
    process::exit,
    thread,
};

#[derive(PartialEq, Eq, Debug)]
enum RequestMethods {
    NULL = -1, // This is only to initialise the struct
    GET,
    HEAD,
}

#[derive(Debug)]
struct RequestLine {
    method: RequestMethods,
    target: String,
    version: String,
}

impl RequestLine {
    pub fn new() -> Self {
        Self {
            method: RequestMethods::NULL,
            target: String::new(),
            version: String::new(),
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc9110#name-methods
    pub fn is_valid_method(method: &str) -> bool {
        if method.trim().is_empty() {
            return false;
        }

        // Only GET and HEAD are required, the rest is optional
        if ["GET", "HEAD"].contains(&method.trim()) {
            return true;
        } else {
            return false;
        }
    }

    // TODO: make the checks less shit and actually correct
    // https://datatracker.ietf.org/doc/html/rfc9112#name-request-target
    pub fn is_valid_target(target: &str) -> bool {
        if target.trim().is_empty() {
            return false;
        }

        // origin-form
        if target.starts_with("/") {
            if target.contains("?") && target.split("?").count() != 2 {
                return false;
            }
            return true;
        }

        // absolute-form
        if target.starts_with("http://") {
            return true;
        }

        // authority-form
        if target.contains(":") && target.split(":").count() == 2 {
            return true;
        }

        // asterisk-form
        if target.trim() == "*" {
            return true;
        }

        return false;
    }

    pub fn is_valid_version(version: &str) -> bool {
        if version.trim().is_empty() {
            return false;
        }

        let http_name = version.trim();

        if http_name.starts_with("HTTP/") {
            let version_numbers = http_name.trim_start_matches("HTTP/");
            let version_numbers = version_numbers.split(".").collect::<Vec<&str>>();
            if version_numbers.len() != 2 {
                return false;
            }
            let major = match version_numbers[0].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };
            let minor = match version_numbers[1].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };

            if major <= 9 && minor <= 9 {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
}

fn parse_start_line(input: &str) -> Result<RequestLine, Vec<u8>> {
    let mut response_field_lines: HashMap<String, String> = HashMap::new();
    let mut response_body: Vec<u8> = vec![];

    if input.ends_with(" ") {
        b"There is whitespace between the start-line and the first field-line"
            .iter()
            .for_each(|byte| response_body.push(*byte));

        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 400 Bad Request",
            response_field_lines,
            Some(response_body),
        ));
    }

    let mut start_line = RequestLine::new();
    let vec = input.trim().split_ascii_whitespace().collect::<Vec<&str>>();

    let body = format!(
        "The start-line has an incorrect amount of items. Got the value: {}",
        vec.len()
    );

    if vec.len() != 3 {
        body.as_bytes()
            .iter()
            .for_each(|byte| response_body.push(*byte));

        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 400 Bad Request",
            response_field_lines,
            Some(response_body),
        ));
    }

    // start_line.method will remain RequestMethods::NULL if it is not supported.
    let method = vec[0];
    if RequestLine::is_valid_method(&method) {
        // TODO: Change to a switch-case if I ever support more methods

        if method == String::from("GET") {
            start_line.method = RequestMethods::GET;
        }
        if method == String::from("HEAD") {
            start_line.method = RequestMethods::HEAD;
        }
    } else {
        b"Server only supports GET and HEAD"
            .iter()
            .for_each(|byte| response_body.push(*byte));

        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 501 Not Implemented",
            response_field_lines,
            Some(response_body),
        ));
    }

    let target = vec[1];
    if RequestLine::is_valid_target(&target) {
        start_line.target = target.to_string();
    }

    let version = vec[2];
    if RequestLine::is_valid_version(&version) {
        if version == "HTTP/1.1" || version == "HTTP/1.0" {
            start_line.version = version.to_string();
        }
    }

    return Ok(start_line);
}

fn parse_field_lines(
    reader: &mut BufReader<&mut TcpStream>,
) -> Result<HashMap<String, String>, Vec<u8>> {
    let mut response_field_lines: HashMap<String, String> = HashMap::new();
    let mut response_body: Vec<u8> = vec![];
    let mut field_lines: HashMap<String, String> = HashMap::new();

    // Read field-lines till I hit an empty line
    loop {
        let mut line = String::new();
        reader.read_line(&mut line).unwrap();

        if !line.ends_with("\r\n") {
            b"Lines need to end with a CRLF"
                .iter()
                .for_each(|byte| response_body.push(*byte));

            response_field_lines.insert(
                String::from("Content-Length"),
                response_body.len().to_string(),
            );
            response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

            return Err(response_builder(
                RequestMethods::GET,
                "HTTP/1.1 400 Bad Request",
                response_field_lines,
                Some(response_body),
            ));
        }

        if line.trim().is_empty() {
            break;
        }

        let field_line = match line.split_once(":") {
            Some(val) => val,
            None => {
                b"Invalid field-line"
                    .iter()
                    .for_each(|byte| response_body.push(*byte));

                response_field_lines.insert(
                    String::from("Content-Length"),
                    response_body.len().to_string(),
                );
                response_field_lines
                    .insert(String::from("Content-Type"), String::from("text/plain"));

                return Err(response_builder(
                    RequestMethods::GET,
                    "HTTP/1.1 400 Bad Request",
                    response_field_lines,
                    Some(response_body),
                ));
            }
        };

        field_lines.insert(field_line.0.to_owned(), field_line.1.trim().to_owned());
    }

    if !field_lines.contains_key(&String::from("Host")) {
        b"field-line with key HOST is missing"
            .iter()
            .for_each(|byte| response_body.push(*byte));

        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 400 Bad Request",
            response_field_lines,
            Some(response_body),
        ));
    }

    return Ok(field_lines);
}

fn response_builder(
    method: RequestMethods,
    status_line: &str,
    field_lines: HashMap<String, String>,
    body: Option<Vec<u8>>,
) -> Vec<u8> {
    let mut response: Vec<u8> = vec![];

    status_line
        .as_bytes()
        .iter()
        .for_each(|byte| response.push(*byte));
    response.push(b'\r');
    response.push(b'\n');

    if !field_lines.is_empty() {
        for field_line in field_lines.iter() {
            field_line
                .0
                .as_bytes()
                .iter()
                .for_each(|byte| response.push(*byte));

            response.push(b':');
            response.push(b' ');

            field_line
                .1
                .as_bytes()
                .iter()
                .for_each(|byte| response.push(*byte));

            response.push(b'\r');
            response.push(b'\n');
        }
    }

    // Mandatory empty line between header and body
    response.push(b'\r');
    response.push(b'\n');

    if method != RequestMethods::HEAD {
        match body {
            Some(val) => {
                val.iter().for_each(|byte| response.push(*byte));
                response.push(b'\r');
                response.push(b'\n');
            }
            None => (),
        }
    }

    return response;
}

fn handle_request(mut stream: TcpStream) -> Result<(), Box<dyn Error>> {
    let mut line = String::new();
    let mut reader = BufReader::new(&mut stream);

    // Request can have one or many empty lines preceding the start-line and I will ignore these
    loop {
        if reader.read_line(&mut line)? > 2 && line != "\r\n" {
            break;
        }
    }

    let start_line = match parse_start_line(&line) {
        Ok(val) => val,
        Err(response) => {
            stream.write_all(&response)?;
            return Ok(());
        }
    };
    dbg!(&start_line);

    let field_lines = match parse_field_lines(&mut reader) {
        Ok(val) => val,
        Err(response) => {
            stream.write_all(&response)?;
            return Ok(());
        }
    };
    dbg!(&field_lines);

    // TODO: Read the body
    // let mut body: Vec<u8> = vec![];
    // reader.read_to_end(&mut body)?;
    // dbg!(&body);

    let mut response_field_lines: HashMap<String, String> = HashMap::new();
    let mut response_body: Vec<u8> = vec![];

    // TODO: Act upon the request
    if start_line.target == "/" {
        let file = match fs::read("./www/index.html") {
            Ok(val) => val,
            Err(_) => {
                b"The is no index.html, only you and me."
                    .iter()
                    .for_each(|byte| response_body.push(*byte));

                response_field_lines.insert(
                    String::from("Content-Length"),
                    response_body.len().to_string(),
                );
                response_field_lines
                    .insert(String::from("Content-Type"), String::from("text/plain"));

                let response = response_builder(
                    start_line.method,
                    "HTTP/1.1 200 OK",
                    response_field_lines,
                    Some(response_body),
                );

                stream.write_all(&response)?;
                return Ok(());
            }
        };

        file.iter().for_each(|byte| response_body.push(*byte));

        response_field_lines.insert(String::from("Content-Length"), file.len().to_string());
        response_field_lines.insert(String::from("Content-Type"), String::from("text/html"));

        let response = response_builder(
            start_line.method,
            "HTTP/1.1 200 OK",
            response_field_lines,
            Some(response_body),
        );

        stream.write_all(&response)?;
    } else {
        let path: PathBuf = PathBuf::from(format!("./www{}", start_line.target));

        match fs::read(path) {
            Ok(val) => {
                val.iter().for_each(|byte| response_body.push(*byte));

                response_field_lines.insert(
                    String::from("Content-Length"),
                    response_body.len().to_string(),
                );
                // TODO: get mime-type of file and use that here
                response_field_lines.insert(String::from("Content-Type"), String::from("*/*"));

                let response = response_builder(
                    start_line.method,
                    "HTTP/1.1 200 OK",
                    response_field_lines,
                    Some(response_body),
                );

                stream.write_all(&response)?;
            }
            Err(_) => {
                let response = response_builder(
                    start_line.method,
                    "HTTP/1.1 404 Not Found",
                    response_field_lines,
                    None,
                );

                stream.write_all(&response)?;
            }
        };
    }

    Ok(())
}

fn main() -> Result<(), Box<dyn Error>> {
    let mut signals = Signals::new([SIGINT, SIGTERM])?;

    // TODO: Gracefully shutdown server
    thread::spawn(move || {
        for sig in signals.forever() {
            println!("Received signal {:?}", sig);
            println!("Shutting down");
            exit(1);
        }
    });

    let listener = TcpListener::bind("0.0.0.0:8080")?;

    println!("Server started");

    for stream in listener.incoming() {
        let stream = stream?;
        handle_request(stream)?;
    }

    Ok(())
}
